/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: main.c,v 1.18.2.1.2.1 2002/02/25 22:20:52 lloydy Exp $
 *
 */
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <netinet/in.h>
#include "chunkyircd.h"
#include "config_file.h"
#include "ircd.h"
#include "logger.h"
#include "network.h"
#include "modules.h"

/*
 * main() - Reads the config file, and the command line parameters
 *	    Logs basic information
 *	    Then calls serve_clients() in ircd.c
 */

/* Used to handle seg faults in modules */
extern int signal_caught;
void quit(int);

int main(int argc,char *argv[])
{
	int server_sockets[32];
	sigset_t newmask,oldmask;
	struct sigaction action;
	int pid_fd=-1;
	char buffer[10];
	int retval=0;
	
	/* Parse Command Line */
	parse_cmdline(argc,argv);
	
	if(shouldfork || !fork()) {
		
		if(!shouldfork) {
			/* Close connections to stdin */
			freopen("/dev/null","r",stdin);
			freopen("/dev/console","w",stdout);
			freopen("/dev/console","w",stderr);
		}
		
		/* Open Log */
		open_log(debug);
		log_msg(LOG_BASIC,"Log opened");
		log_msg(LOG_BASIC,"Chunky Monkey IRCD, Version %s",VERSION);
		log_msg(LOG_BASIC,"Built at %s on %s",__TIME__,__DATE__);

		/* Init the modules, and load the config file */
		ircd_module_init();
		parse_config_file("ircd.conf",server_sockets);

#ifdef USE_OPENSSL
		/* Init Openssl */
		log_msg(LOG_BASIC,"Initialising OpenSSL layer");
		OpenSSL_add_all_algorithms();
#endif
		
		/* Output PID to /tmp/monkeyircd.pid */
		pid_fd=open("/tmp/monkeyircd.pid",O_WRONLY ^ O_CREAT);
		if(pid_fd==-1) {
			log_msg(LOG_BASIC,"Unable to write PID (%d) to /tmp/monkeyircd.pid",getpid());
		}
		sprintf(buffer,"%d\n",getpid());
		write(pid_fd,buffer,strlen(buffer));
		close(pid_fd);
			
		/* Call to get a socket */
		if(server_sockets[0]!=-1) {
			/* Setup signal handling, then go */
			sigemptyset(&newmask);
			
			/* If We're attached to a tty, don't mask SIGINT */
			if(!shouldfork) 
				sigaddset(&newmask,SIGINT);
			
			/* ignore Broken Pipe errors, deal with them ourselves */
			sigaddset(&newmask,SIGPIPE);
			sigprocmask(SIG_BLOCK,&newmask,&oldmask);
			
			/* Make SIGTERM exit us gracefully */
			action.sa_flags=0;
			action.sa_handler=quit;
			sigaction(SIGTERM,&action,NULL);

			/* Rehash on a SIGHUP */
			action.sa_flags=0;
			action.sa_handler=rehash;
			sigaction(SIGHUP,&action,NULL);
			
			/* Exit nicely with SIGINIT if we're attached to a TTY */
			if(!shouldfork) {
				action.sa_handler=quit;
				action.sa_flags=0;
				sigaction(SIGINT,&action,NULL);
			}

			/* Handle segfaults in modules */
			action.sa_flags=0;
			action.sa_handler=caught_segfault;
			sigaction(SIGSEGV,&action,NULL);

			/* Start serving */
			serve_clients(server_sockets);
			
			/* Close the Log File and finish with OpenSSL */
#ifdef USE_OPENSSL
			log_msg(LOG_DEBUG1,"Cleaning up OpenSSL");
			EVP_cleanup();
#endif
			log_msg(LOG_BASIC,"Log file closed");
			close_log();
			retval=0;
		} else {
			log_msg(LOG_BASIC,"Failed to bind to any sockets\n");
			close_log();
			retval=1;
		}
		
		/* get rid of pid file */
		unlink("/tmp/monkeyircd.pid");
	}

	return retval;
}

/* 
 * Called when we get a signal
 * Basically sets a flag, and we exit cleanly
 */
void quit(int signal)
{
	if(signal == SIGINT || signal == SIGTERM)
		signal_caught=1;
}

