/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: network.c,v 1.31 2002/02/19 15:22:13 lloydy Exp $
 *
 */
#include "network.h"

/* Utility Routines */
static int send_server(ServerInfo ,void *,int length);
static int send_raw(int,void *,int length);

/* 
* called to setup tcp socket, returns socket or -1
*/
int setup_tcp(struct sockaddr_in * server)
{
	int tcp_socket,sock_val;
	
	/* Get TCP Socket */
	if((tcp_socket=socket(AF_INET,SOCK_STREAM,0))==-1) {
		perror("socket");
		return -1;
	}

	/* Set reuse */
	sock_val=1;
	setsockopt(tcp_socket,SOL_SOCKET,SO_REUSEADDR,&sock_val,sizeof(int));

	/* Bind to socket */
	if(bind(tcp_socket,(struct sockaddr *)server,sizeof(struct sockaddr))==-1) {
		close(tcp_socket);
		perror("bind");
		return -1;
	}
	
	/* Listen on socket */
	if(listen(tcp_socket,NET_MAXLISTEN)==-1) {
		close(tcp_socket);
		perror("listen");
		return -1;
	}
		
	/* Set non-blocking */
	fcntl(tcp_socket,F_SETFL,O_NONBLOCK);
	
	log_msg(LOG_DEBUG1,"Successfully bound to %s:%d",inet_ntoa(server->sin_addr),ntohs(server->sin_port));
	
	return tcp_socket;

}
/* 
* send a message to a client, add the necessary stuff depending
* on whether they are local or not 
*/
int send_client_client(ClientInfo dest_ptr,ClientInfo src_ptr,char *message)
{
	char buffer[BUF_SIZE];
	/* message is NICK timmy, or suchlike */

	if(dest_ptr->fd!=-1) {
		/* Local recipient */
		if(src_ptr) {
			ircd_sprintf(buffer,BUF_SIZE,":%s!%s@%s %s\n",src_ptr->nick,
				    src_ptr->user,src_ptr->host,message);
		} else {
			ircd_sprintf(buffer,BUF_SIZE,"%s\n",message);
		}
		return send_raw(dest_ptr->fd,buffer,strlen(buffer));
	} else {
		ircd_sprintf(buffer,BUF_SIZE,":%s %s\n",src_ptr->nick,message);
		return send_server(dest_ptr->server,buffer,strlen(buffer));
	}
}

/*
*  Send a client messages to a server
*/
int send_origin_server(ServerInfo dest,char *origin,char *message)
{
	unsigned char buffer[512];

	ircd_sprintf( (char *) buffer,512,"%s%s%s%s\n",origin?":":"",origin?origin:"",origin?" ":"",message);
	
#ifdef USE_OPENSSL
	if(dest->ctx) {
		unsigned char enc_buffer[512];
		int enc_length;
		/* Encrypt */
		encrypt_buffer(dest->ctx,buffer,enc_buffer,&enc_length);
		/* Send the length */
		ircd_sprintf( (char *) buffer,5,"%04d",enc_length);
		buffer[4]='\0';
		send_server(dest, (char *) buffer,4);
		/* Send the encrypted data */
		return send_server(dest, (void *) enc_buffer,enc_length);
	} else {
		return send_server(dest,(void *) buffer,strlen( (char *) buffer));
	}
#else
	return send_server(dest,buffer,strlen((char *)buffer));
#endif
}

/*
* Send something to a temp
*/
int send_temp(struct tempinfo *dest,char *message)
{
	unsigned char buffer[512];
	
	ircd_sprintf( (char *) buffer,512,":%s %s\n",server_name,message);
#ifdef USE_OPENSSL
	if(dest->ctx) {
		unsigned char enc_buffer[512];
		int enc_length;
		/* Encrypt */
		encrypt_buffer(dest->ctx,buffer,enc_buffer,&enc_length);
		/* Send the Length */
		ircd_sprintf( (char *) buffer,5,"%04d",enc_length);
		buffer[4]='\0';
		send_raw(dest->fd,buffer,4);
		/* Send the encrypted data */
		return send_raw(dest->fd,enc_buffer,enc_length);
	} else {
		return send_raw(dest->fd,buffer,strlen( (char *) buffer));
	}
#else
	return send_raw(dest->fd,buffer,strlen((char *)buffer));
#endif
	
}

/* 
* Send origin only to client
*/
int send_origin_client(ClientInfo dest,char *origin,char *message)
{
	char buffer[512];

	ircd_sprintf(buffer,512,"%s%s%s%s\n",origin?":":"",origin?origin:"",origin?" ":"",message);

	if(dest->fd==-1) {
		return send_server(dest->server,buffer,strlen(buffer));
	} else {
		return send_raw(dest->fd,buffer,strlen(buffer));
	}
	
}

/* 
* Sends raw output to a socket 
*/
static int send_raw(int fd,void *buffer,int length)
{
/*	log_msg(LOG_INFO,"Sent: %s",buffer); */
	return send(fd,buffer,length,0);
}


/* 
* Send to a server, recursive 
*/
static int send_server(ServerInfo s_ptr,void *buffer,int length)
{
	if(s_ptr->fd==-1) {
		return send_server(s_ptr->via_ptr,buffer,length);
	} else {
		return send_raw(s_ptr->fd,buffer,length);
	}

}

/* 
* sends a numeric reply 
* only local clients can get this sent to them 
*/
int send_numeric(ClientInfo dest,ServerInfo src,int code,char *text)
{
	char buffer[480];
	
	ircd_sprintf(buffer,480,"%03d %s %s",code,dest->nick,text);
	
	return send_origin_client(dest,src->name,buffer);
}

/* 
* Loop though a channel and send a message 
* Sends to clients on other servers
*/
int send_to_channel(ClientInfo sender,ChannelInfo chan,char *message)
{
	ChannelMember curr_client=NULL;
	
	curr_client=chan->users_start;
	
	while(curr_client) {
		if(curr_client->user_info!=sender) {
			/* Clients on another server are dealt with by that server */
			send_client_client(curr_client->user_info,sender,message);
		}
		curr_client=curr_client->next;
	}

	return 1;
}

/* 
* Only send to local channel members
*/
int send_to_local_channel(ClientInfo sender,ChannelInfo chan,char *message)
{
	ChannelMember curr_client=NULL;
	
	curr_client=chan->users_start;
	
	while(curr_client) {
		if(curr_client->user_info!=sender && curr_client->user_info->fd!=-1) {
			/* Clients on another server are dealt with by that server */
			send_client_client(curr_client->user_info,sender,message);
		}
		curr_client=curr_client->next;
	}

	return 1;
}

/*
* Send to local channel ops
*/
int send_to_local_channel_ops(ClientInfo sender,ChannelInfo chan,char *message)
{
	ChannelMember curr_client=NULL;

	curr_client=chan->users_start;

	while(curr_client) {
		if(curr_client->user_info!=sender && curr_client->user_info->fd!=-1 && (curr_client->chan_mode & CHANUSR_OP) ) {
			send_client_client(curr_client->user_info,sender,message);
		}
		curr_client=curr_client->next;
	} 

	return 1;
}

/* Non-basic routines - all depend on routines above */
/* Basically, they could be in cmd_*.c, but due to the */
/* dynamic stuff, and the fact other routines rely on them */
/* they are in here */

/* Send MOTD to a client */
int send_motd(ClientInfo c_ptr)
{
	char buffer[BUF_SIZE];
	int i;
	ServerInfo m_ptr=NULL;

	m_ptr=get_my_server();
	
	if(motd==NULL) {
		/* Send a numeric to tell them */
		send_numeric(c_ptr,m_ptr,422,":MOTD File is missing");
	} else {
		/* Send the motd */
		ircd_sprintf(buffer,BUF_SIZE,":- %s Message of the day -",server_name);
		send_numeric(c_ptr,m_ptr,375,buffer);
		
		for(i=0;motd[i]!=NULL;i++) {
			ircd_sprintf(buffer,BUF_SIZE,":- %s",motd[i]);
			send_numeric(c_ptr,m_ptr,372,buffer);
		}
		
		send_numeric(c_ptr,m_ptr,376,":End of MOTD Command");
	}

	return 1;
}

/* 
* Send a ping
*/
int send_ping(int fd,char *target)
{
	char buffer[200];
	ircd_sprintf(buffer,200,"PING %s%s%s\n",server_name,target?" :":"",target?target:"");
	send_raw(fd,buffer,strlen(buffer));

	return 1;
}

/* 
* Send a channel name list to a client 
*/
int send_namelists(ChannelInfo chan, ClientInfo  client)
{
	char start_char='=';
	char buffer[BUF_SIZE];
	char next[BUF_SIZE];
	char nicks[BUF_SIZE];
	char start[BUF_SIZE];
	ChannelMember curr_client=NULL;
	ChannelMember on_chan=NULL;
	unsigned int avail_chars=0;
	
	/* Check to see if person who called is on channel */
	on_chan=channel_is_client_on(client,chan);
	
	/* decide on start char */
	switch(chan->mode & (CMODE_SECRET | CMODE_PRIVATE) ) {

		case (CMODE_SECRET):
			start_char='@';
			break;

		case (CMODE_PRIVATE):
			start_char='*';
			break;

		default:
			start_char='=';
			break;
	}

	/* Start to formulate list */
	ircd_sprintf(start,BUF_SIZE,"%c %s :",start_char,chan->name);
	avail_chars=BUF_SIZE - 1 - strlen(start);
	nicks[0]='\0';
	
	/* Sweep through channel */
	if(!(chan->mode & CMODE_SECRET) || (on_chan)) {
		curr_client=chan->users_start;	
	
		while(curr_client!=NULL) {
			
			if( !(curr_client->user_info->mode & MODE_INVIS) || (on_chan)) {

				ircd_sprintf(next,BUF_SIZE,"%s%s ",
							(curr_client->chan_mode & CHANUSR_OP)?"@":"",
					    		curr_client->user_info->nick);
				if(strlen(next) < avail_chars) {
					strcat(nicks,next);
					avail_chars -= strlen(next);
				} else {
					ircd_sprintf(buffer,BUF_SIZE,"%s%s",start,nicks);
					send_numeric(client,get_my_server(),353,buffer);
					avail_chars=BUF_SIZE - 1 - strlen(start) - strlen(next);
					strcpy(nicks,next);
				}
					
			}

			curr_client=curr_client->next;
		}
		
		/* Send list */
		ircd_sprintf(buffer,BUF_SIZE,"%s%s",start,nicks);
		send_numeric(client,get_my_server(),353,buffer);
	}

	return 1;
}

/*
* Send a LUSERS set of replies when prompted
* by a LUSERS command or at client registration
*/
int send_lusers(ClientInfo c_ptr,char *mask) 
{
	char buffer[BUF_SIZE];
	ServerInfo my_server=NULL;
	long int global_clients;
	long int global_servers;

	/* Get stuff we need */
	my_server=get_my_server();
	servers_global_counts(mask,&global_clients,&global_servers);
	
	/* Send LUSERCLIENT */
	ircd_sprintf(buffer,BUF_SIZE,":There are %ld users and %0d services on %ld servers",
								global_clients,
								0,
								global_servers);
	send_numeric(c_ptr,my_server,251,buffer);

	/* Send LUSEROP */
	if(ops) {
		ircd_sprintf(buffer,BUF_SIZE,"%0ld :operators online",ops);
		send_numeric(c_ptr,my_server,252,buffer);
	}

	/* Send LUSERUNKNOWN */
	if(temp_count) {
		ircd_sprintf(buffer,BUF_SIZE,"%ld :unknown clients",temp_count);
		send_numeric(c_ptr,my_server,253,buffer);
	}

	/* Send LUSERCHANNELS */
	if(channels) {
		ircd_sprintf(buffer,BUF_SIZE,"%ld :channels formed",channels);
		send_numeric(c_ptr,my_server,254,buffer);
	}

	/* Send LUSERME */
	ircd_sprintf(buffer,BUF_SIZE,":I Have %ld clients and %ld servers",my_server->client_count,my_server->server_count);
	send_numeric(c_ptr,my_server,255,buffer);

	return 1;
}

/* 
* Force a peer to disconnect 
*/
int force_quit(int fd,char *reason)
{
	char buffer[BUF_SIZE];
	int i;
	
	if(fd_list[fd].temp) {
		ircd_sprintf(buffer,BUF_SIZE,"ERROR :Closing Link :%s",reason);
		log_msg(LOG_DEBUG1,"Closed Temporary link - %s",reason);
		send_temp(fd_list[fd].temp,buffer);
		fd_list[fd].temp=NULL;
		/* Just dequeue and close port */
		temp_rem(fd);
		close(fd);
		return 1;
	}

	if(fd_list[fd].client) {
		/* Force kill client */
		ircd_sprintf(buffer,BUF_SIZE,"QUIT :%s",reason);

		/* XXX - May Send Multiple times */
		if(fd_list[fd].client->chan_list!=NULL) {
			for(i=0;i<fd_list[fd].client->chanlist_no && fd_list[fd].client->chan_list[i]!=NULL;i++) {
				/* remove from channel user list */
				channel_rem_client(fd_list[fd].client,fd_list[fd].client->chan_list[i]);
				
				/* Infor	m channel */
				send_to_channel(fd_list[fd].client,fd_list[fd].client->chan_list[i],buffer);
			
				/* Close channel if needed */
				channel_needclose(fd_list[fd].client->chan_list[i]);

				/* Set pointer to NULL */
				fd_list[fd].client->chan_list[i]=NULL;
			}
		}
		
		/* Announce to Servers */
		servers_announce(fd_list[fd].client->nick,buffer);

		if(fd==-1) {
			if(fd == highest_fd) {
				for(i=fd;i>=0 ;i--) {
					if(fd_list[fd].client) {
						highest_fd=i;
						i=-1;
					}
				}
				if(i==0) 
					highest_fd=-1;
			}
		}
		
		/* Report we're closing the link */
		ircd_sprintf(buffer,BUF_SIZE,"ERROR :Closing Link :%s",reason);
		send_origin_client(fd_list[fd].client,NULL,buffer);
		
		
		/* Remove from tree */
		client_remove(fd_list[fd].client->nick);

		if((fd_list[fd].client->mode & MODE_IRCOP))
			ops--;
		get_my_server()->client_count--;
		
		/* Remove from fd list */
		if(fd!=-1)  {

			fd_list[fd].client=NULL;
			close(fd);
			
		}

		return 1;
	}

	if(fd_list[fd].server) {
		/* Log it */
		log_msg(LOG_BASIC,"Broke Local Server link to %s. Reason: %s",fd_list[fd].server->name,reason);
		
		if(fd == highest_fd) {
			for(i=fd;i>=0 ;i--) {
				if(fd_list[fd].server) {
					highest_fd=i;
					i=-1;
				}
			}
			if(i==0) 
				highest_fd=-1;
		}

		/* Inform other servers */
		server_squit_announce(fd_list[fd].server,reason);
		
		/* Work out clients on that server */
		clients_netsplit(fd_list[fd].server,fd_list[fd].server->name);
		
		/* Work out which clients and servers behind this one where disconnected */
		server_squit_generate(fd_list[fd].server,reason,fd_list[fd].server->name);

		/* Tell other servers we netsplitted from this server, and they work out the rest */
		ircd_sprintf(buffer,BUF_SIZE,"SQUIT %s :%s",fd_list[fd].server,reason);
		servers_relay(fd_list[fd].server,get_my_server()->name,buffer);
			
		rem_server_hash(fd_list[fd].server->name);
		get_my_server()->server_count--;

		free(fd_list[fd].server);
		
		fd_list[fd].server=NULL;

		close(fd);

		return 1;
		
	
	}

	return 0;
		
}

