/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: server.c,v 1.18 2002/01/02 15:58:20 lloydy Exp $
 *
 */
#include "server.h"

/* server.c - routines related to server activites
 * mainly contains code to add/remove servers to the
 * hash table. */


/* Actual table */
static ServerInfo servers[SERV_HASH_SIZE];
/* Pointer to our structure */
static ServerInfo my_ptr=NULL;

/* Routines */
static int get_subscript_by_name(char *);

/* initialise hash table, and add us */
int init_server_hash(char *my_name,char *my_info) 
{
	int i;
	ServerInfo my_serv=NULL;
	
	/* NULL out hash */
	for(i=0;i<SERV_HASH_SIZE;i++) {
		servers[i]=NULL;
	}

	/* Attempt to alloc memory for our info */
	my_serv=ircd_malloc(sizeof(struct _serverinfo));

	
	/* Setup Struct */
	strncpy(my_serv->name,my_name,SERVNAME_MAX);
	strncpy(my_serv->info,my_info,SERVNAME_MAX);
	my_serv->fd=-1;
	my_serv->hops=0;
	my_serv->via_ptr=NULL;

	/* Add to hash */
	if(!add_server_hash(my_serv)) {
		printf("Created my info, but failed to add it to the hash\n");
	} 
	my_ptr=my_serv;

	return 1;

}
	

/* Add a server into the hash table */
int add_server_hash(ServerInfo server) 
{
	int subscript,i;

	/* Get inital subscript */
	subscript=get_subscript_by_name(server->name);	
	
	/* Find a free one, if this isn't free */
	for(i=1;servers[subscript]!=NULL;i=2*i) {
		/* Move along */
		subscript += i;
		/* Reduce if too large */
		if(subscript>=SERV_HASH_SIZE)
			subscript-=SERV_HASH_SIZE;

	}
	
		
	/* Add in */
	servers[subscript]=server;
	
	return 1;

}

/* Remove a table from the hash */
/* Don't free() it */
int rem_server_hash(char *name) 
{
	int subscript,i;

	subscript=get_subscript_by_name(name);
	
	
	for(i=1;servers[subscript]!=NULL && lc_strncmp(name,servers[subscript]->name,-1)!=0;i=2*i) {
		subscript += i;
		if(subscript>=SERV_HASH_SIZE)
			subscript -= SERV_HASH_SIZE;
	}

	if(servers[subscript]==NULL) {
		/* Couldn't find it */
		return EXIT_FAILURE;
	} else {
		servers[subscript]=NULL;
		return EXIT_SUCCESS;
	}
}	

/* Locate and return a pointer to the server with a name which matches this one */
/* return NULL on failure */
ServerInfo locate_server(char *name) 
{
	int subscript,i;

	subscript=get_subscript_by_name(name);
	
	for(i=1;servers[subscript]!=NULL && lc_strncmp(name,servers[subscript]->name,-1)!=0;i=2*i) {
		subscript += i;
		if(subscript>=SERV_HASH_SIZE)
			subscript -= SERV_HASH_SIZE;
	}

	if(servers[subscript]==NULL) {
		return NULL;
	} else {
		return servers[subscript];
	}
}

/* Get a subscript for the server name */
static int get_subscript_by_name(char *name)
{
	int subscript,i,total=0;
	
	for(i=0;name[i]!='\0';i++) {
		if(name[i]<96) {
			total+= 2 * (name[i]+32);
		} else {
			total+= 2 * name[i];
		}
	}

	subscript = total % SERV_HASH_SIZE;
	
	return subscript;
}


/* rest of these commands are used to inform servers of occurances on this server */
/* 
 * Inform servers connected directly to us 
 * Takes a char * param to show who sent it
 * Usually used for events occuring on this server 
 */
int servers_announce(char *origin,char *text)
{
	int i;
	
	for(i=0;i<SERV_HASH_SIZE;i++) {
		if(servers[i] && (servers[i]->fd!=-1) && servers[i]!=my_ptr ) {
			/* send */
			send_origin_server(servers[i],origin,text);
		}
	}

	return 1;
}

/*
 * Inform servers connected to us
 * used for events not occuring on this server
 */
int servers_relay(ServerInfo source,char *origin,char *text)
{
	int i;

	for(i=0;i<SERV_HASH_SIZE;i++) {
		if(servers[i] && servers[i]->fd!=-1 && servers[i]!=my_ptr && servers[i]!=source) {
			/* Send */
			send_origin_server(servers[i],origin,text);
		}
	}

	return 1;
}

/*
* Introduce a particular server
*/
static int introduce_server(ServerInfo recip,long int rand_no,ServerInfo intro)
{
	char buffer[BUF_SIZE];
	
	if(intro->rand_no==rand_no)  {
		return 0;
	} else {
		if(intro->via_ptr->rand_no!=rand_no) {
			introduce_server(recip,rand_no,intro->via_ptr);
		}
		sprintf(buffer,"SERVER %s %0d 0 :%s",intro->name,intro->hops+1,
						       intro->info);
		send_origin_server(recip,intro->via_ptr->name,buffer);
		intro->rand_no=rand_no;
		return 1;
	}
}

/* 
 * Dump servers to a newly registered server
 * Don't send the server we're sending it do 
 */
int servers_introduce(ServerInfo recip)
{
	int i;
	long int rand_no;

	rand_no=rand() % 1000 + 1;
	get_my_server()->rand_no=rand_no;
	
	if(recip->fd==-1) {
		return 0;
	}
	
	for(i=0; i < SERV_HASH_SIZE;i++) {
		if(servers[i] && servers[i]!=recip && servers[i]!=my_ptr) {
			/* Send them the info */
			if(servers[i]->rand_no!=rand_no) {
				introduce_server(recip,rand_no,servers[i]);
			}
		}
	}
	
	return 1;
}

/* Announce an SQUIT, for whatever reason */
int server_squit_announce(ServerInfo s_ptr,char *reason)
{
	char buffer[BUF_SIZE];
	
	ircd_sprintf(buffer,BUF_SIZE,"Got a server quit for %s.  Reason: %s",s_ptr->name,reason);
	clients_servnotice(buffer);
		
	return 1;
}

/* 
* Generate squits for servers connected via ours 
* s_ptr is server which died
* any server with via_ptr==s_ptr, should be killed 
*/
int server_squit_generate(ServerInfo s_ptr,char *reason,char *org_server_name) 
{
	int i;
	char buffer[BUF_SIZE];
	ServerInfo temp=NULL;
	
	ircd_sprintf(buffer,BUF_SIZE,"%s:%s",s_ptr->name,reason);
	for(i=0; i < SERV_HASH_SIZE ;i++) {
		if(servers[i] && servers[i]!=s_ptr && servers[i]->via_ptr==s_ptr && servers[i]!=my_ptr) {

			/* Send out SQUIT message */
			server_squit_announce(servers[i],buffer);
			
			/* Check for clients below this one */
			server_squit_generate(servers[i],reason,org_server_name);
			
			/* Work out which clients are dead */
			clients_netsplit(servers[i],org_server_name);

			/* Remove from hash */
			temp=servers[i];
			rem_server_hash(servers[i]->name);
			temp->via_ptr->server_count--;
			servers[i]=NULL;
			
			/* Free() it */
			free(temp);
		}

	}

	return 1;
}

/* 
* Send the LINKS replies to the client 
*/
int server_links(ClientInfo c_ptr,char *mask) 
{
	int i;
	char buffer[BUF_SIZE];

	if(c_ptr->fd==-1) 
		return 0;

	for(i=0; i < SERV_HASH_SIZE ;i++) {
		if(servers[i] && check_string(servers[i]->name,mask)) {
			/* It isn't us, and it matches the mask */
			ircd_sprintf(buffer,BUF_SIZE,"%s %s :%ld %s",servers[i]->name,
						servers[i]==my_ptr?servers[i]->name:servers[i]->via_ptr->name,
								     servers[i]->hops,servers[i]->info);

			send_numeric(c_ptr,get_my_server(),364,buffer);
		}
	}

	return 1;
}
	

/* 
* Return pointer to our server struct in memory 
*/
ServerInfo get_my_server()
{
	return my_ptr;
}

/*
* Generate LUSERS for servers which match a mask passed
*/
int servers_global_counts(char *mask,long int *client_count, long int *server_count)
{	
	int i;

	/* Set counters to 0 */
	(*client_count)=0;
	(*server_count)=1;

	for(i=0 ; i < SERV_HASH_SIZE ; i++) {
		if( servers[i]!=NULL && check_string(servers[i]->name,mask)!=0 ) {
			(*client_count) += servers[i]->client_count;
			(*server_count) += servers[i]->server_count;
		}
	}

	return 1;
}
